from .read import ReadClient
from .write import WriteClient
from .queue import QueueClient
from .request import Request, Response, bound_request
from .pipeline import PipelineClients, clients, input_client, output_client, queue_clients

__all__ = [
  'ReadClient', 'WriteClient', 'QueueClient',
  'Request', 'Response', 'bound_request',
  'PipelineClients', 'clients', 'input_client', 'output_client', 'queue_clients'
]


from typing import TypeVar, Generic, overload, Callable
from pydantic import RootModel
from haskellian import Either, Right, either as E
from pipeteer.queues import Queue, QueueError
from .read import ReadClient
from .write import WriteClient
from .request import request, Request
from .util import urljoin

T = TypeVar('T')

class QueueClient(Queue[T], ReadClient[T], WriteClient[T], Generic[T]):

  @overload
  def __init__(
    self, url: str, *,
    parse: Callable[[bytes], Either[QueueError, T]] = Right,
    dump: Callable[[T], bytes] = lambda x: x, # type: ignore
    request: Request = request
  ): ...
  @overload
  def __init__(
    self, *, read_url: str, write_url: str,
    parse: Callable[[bytes], Either[QueueError, T]] = Right,
    dump: Callable[[T], bytes] = lambda x: x, # type: ignore
    request: Request = request
  ): ...
  def __init__(
    self, url = None, *, read_url = None, write_url = None,
    parse: Callable[[bytes], Either[QueueError, T]] = Right, dump = lambda x: x,
    request: Request = request
  ):
    if url:
      self.read_url = urljoin(url, 'read')
      self.write_url = urljoin(url, 'write')
    elif read_url and write_url:
      self.read_url = read_url
      self.write_url = write_url
    else:
      raise ValueError('Either url or (read_url and write_url) must be provided')

    ReadClient.__init__(self, self.read_url, parse=parse, request=request)
    WriteClient.__init__(self, self.write_url, dump=dump, request=request)


  @overload
  @classmethod
  def validated(cls, Type: type[T], url: str, *, request: Request = request) -> 'QueueClient[T]': ...
  @overload
  @classmethod
  def validated(cls, Type: type[T], *, read_url: str, write_url: str, request: Request = request,) -> 'QueueClient[T]': ...
  @classmethod
  def validated(cls, Type, url = None, *, read_url = None, write_url = None, request = request) -> 'QueueClient[T]': # type: ignore
    Model = RootModel[Type]
    return cls(
      url=url, read_url=read_url, write_url=write_url, request=request,
      parse=lambda b: E.validate_json(b, Model).fmap(lambda x: x.root).mapl(QueueError),
      dump=lambda x: Model(x).model_dump_json(exclude_none=True)
    ) # type: ignore


from typing_extensions import TypeVar, Generic, Sequence, Callable, Awaitable, Any
from dataclasses import dataclass
import asyncio
from haskellian import Either, Left, Right, either as E, promise as P
from pipeteer.queues import WriteQueue, QueueError, ReadQueue

A = TypeVar('A')
B = TypeVar('B')
S1 = TypeVar('S1')
S2 = TypeVar('S2')

@dataclass
class prejoin(WriteQueue[A], Generic[A]):
  """Pre-join queues based on the pushed value type."""

  outputs: Sequence[tuple[WriteQueue[A], type[A]]]
  fallback: WriteQueue[A] | None = None

  def __repr__(self):
    reprs = ', '.join(repr(q) for q, _ in self.outputs)
    return f'prejoin({reprs})'

  async def push(self, key: str, value: A) -> Either[QueueError, None]:
    for q, t in self.outputs:
      if isinstance(value, t):
        return await q.push(key, value)
    if not self.fallback:
      return Left(QueueError(f'Received value with type {type(value)}, which doesn\'t match any of the available types. Value: {value}'))
    return await self.fallback.push(key, value)

@dataclass
class premerge(WriteQueue[B], Generic[B, S1, S2]):
  """Merges an input queue `Qin` with the pushed values using `merge`"""
  
  Qin: ReadQueue[S1]
  Qout: WriteQueue[S2]
  merge: Callable[[S1, B], Awaitable[S2] | S2]

  def __repr__(self):
    return f'premerge({repr(self.Qin)} -> {repr(self.Qout)})'

  @E.do[QueueError]()
  async def push(self, key: str, value: B): # type: ignore (python, bruh)
    state = (await self.Qin.read(key)).unsafe()
    next = await P.wait(self.merge(state, value))
    (await self.Qout.push(key, next)).unsafe()
    (await self.Qin.pop(key)).unsafe()


class tee(WriteQueue[A], Generic[A]):
  """A queue that pushes to the multiple queues at once"""

  def __init__(self, q1: WriteQueue[A], q2: WriteQueue[A], /, *qs: WriteQueue[A]):
    self.queues: Sequence[WriteQueue[A]] = [q1, q2, *qs]

  def __repr__(self) -> str:
    reprs = ', '.join(repr(q) for q in self.queues)
    return f'tee({reprs})'

  @E.do[QueueError]()
  async def push(self, key: str, value: A): # type: ignore (duh)
    results = await asyncio.gather(*[q.push(key, value) for q in self.queues])
    E.sequence(results).mapl(QueueError).unsafe()


class prefilter(WriteQueue[A], Generic[A]):
  
  def __init__(self, q: WriteQueue[A], p: Callable[[tuple[str, A]], Awaitable[bool]]):
    self._wrapped = q
    self._predicate = p

  def __repr__(self):
    return f'prefilter({self._wrapped!r})'

  async def push(self, key: str, value: A) -> Either[QueueError, None]:
    if await self._predicate((key, value)):
      return await self._wrapped.push(key, value)
    return Right(None)


class premap(WriteQueue[B], Generic[A, B]):
  
  def __init__(self, q: WriteQueue[A], f: Callable[[tuple[str, B]], Awaitable[Either[Any, tuple[str, A]]]]):
    self._wrapped = q
    self._mapper = f

  def __repr__(self):
    return f'premap({self._wrapped!r})'

  @E.do()
  async def push(self, key: str, value: B):
    k, v = (await self._mapper((key, value))).unsafe()
    return (await self._wrapped.push(k, v)).unsafe()


from types import UnionType
from .pipeline import Pipeline

def indent(s: str, n: int = 1):
  return '\n'.join(' ' * n + line for line in s.splitlines())

def name(type):
  if isinstance(type, UnionType):
    return 'Union'
  return getattr(type, '__name__', str(type))

def str_types(pipeline: Pipeline):
  types = name(pipeline.Tin)
  if pipeline.Tout:
    types += f' -> {name(pipeline.Tout)}' # type: ignore
  return types


from typing import Any, TypeVar
from fastapi import FastAPI
from pipeteer import http, trees, ReadQueue, GetQueue, Pipeline

T = TypeVar('T')
P = TypeVar('P')

def mount(pipeline: Pipeline[Any, T, Any, P, Any], Qout: ReadQueue[T], get_queue: GetQueue, params: P) -> FastAPI:
  """Mounts the pipeline's queues into a FastAPI app
  - `/input/write`: write API for the input queue
  - `/{path/to/queue}/read`: read API for each observable queue
  - `/output/read`: read API for the output queue
  """
  api = FastAPI()
  Qpush = pipeline.push_queue(get_queue, params)
  Qs = pipeline.observe(get_queue)
  api.mount('/input/write', http.write_api(Qpush, Type=pipeline.Tin))
  Tout: type = pipeline.Tout or Any # type: ignore (duh)
  api.mount('/output/read', http.read_api(Qout, Type=Tout))
  
  for path, queue in trees.flatten(Qs):
    AnyT: type = Any # type: ignore (duh)
    slug = '/' + '/'.join(str(p) for p in path) + '/read'
    api.mount(slug, http.read_api(queue, Type=AnyT))

  return api


from .queues import ReadQueue, WriteQueue, Queue, SimpleQueue, QueueKV
from .pipelines import Pipeline, Wrapped, Workflow, Task, GetQueue, Sequential
from . import trees, http

__all__ = [
  'ReadQueue', 'WriteQueue', 'Queue', 'SimpleQueue', 'QueueKV',
  'Pipeline', 'Wrapped', 'Workflow', 'Task', 'GetQueue', 'Sequential',
  'trees', 'http',
]


from typing import TypeVar, Generic, TypedDict, ParamSpec
from dataclasses import dataclass
from abc import abstractmethod
from pipeteer.queues import ReadQueue, WriteQueue
from pipeteer import Pipeline, GetQueue, trees

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
T = TypeVar('T')
P = TypeVar('P')


class TaskQueues(TypedDict, Generic[A, B]):
  Qin: ReadQueue[A]
  Qout: WriteQueue[B]

class Task(Pipeline[A, B, TaskQueues[A, B], P, T], Generic[A, B, P, T]):
  """A pipeline that reads from a single input queue, writes to a single output queue"""

  Queues = TaskQueues

  def __repr__(self):
    from .reprs import str_types
    return f'Task[{str_types(self)}]'

  def push_queue(self, get_queue: GetQueue, params: P, /, *, prefix: tuple[str | int, ...] = ()) -> WriteQueue[A]:
    return get_queue(prefix, self.Tin)
  
  def connect(self, Qout: WriteQueue[B], get_queue: GetQueue, params: P, /, *, prefix: tuple[str|int, ...] = ()):
    Qin = get_queue(prefix, self.Tin)
    return TaskQueues(Qin=Qin, Qout=Qout)

  def tree(self) -> trees.Tree[Pipeline]:
    return self
  
  @abstractmethod
  def run(self, queues: TaskQueues[A, B], params: P, /) -> T:
    ...
    


from .client import QueueClient, ReadClient, WriteClient, bound_request, clients, PipelineClients
from .server import queue_api, read_api, write_api, mount
from . import client, server

__all__ = [
  'QueueClient', 'ReadClient', 'WriteClient', 'bound_request',
  'clients', 'PipelineClients',
  'queue_api', 'read_api', 'write_api', 'mount',
  'client', 'server',
]


from typing import TypeVar, Generic, Callable, Literal
from haskellian import Either, Left, Right
from pipeteer.queues import WriteQueue, QueueError
from .util import validate_left, urljoin
from .request import request, Request

T = TypeVar('T')

class WriteClient(WriteQueue[T], Generic[T]):

  def __init__(
    self, url: str, *,
    dump: Callable[[T], bytes] = lambda x: x, # type: ignore
    request: Request = request
  ):
    self.write_url = url
    self.dump = dump
    self.request = request

  @classmethod
  def validated(cls, Type: type[T], url: str, *, request: Request = request) -> 'WriteClient[T]':
    from pydantic import TypeAdapter
    dump = TypeAdapter(Type).dump_json
    return cls(url, dump=dump, request=request)


  async def _write_req(
      self, method: Literal['GET', 'POST', 'DELETE'], path: str, *,
      data: bytes | str | None = None, params: dict = {}
    ):
      try:
        r = await self.request(method, urljoin(self.write_url, path), data=data, params=params)
        return Right(r.content) if r.status_code == 200 else validate_left(r.content, r.status_code)
      except Exception as e:
        return Left(QueueError(e))
      
  async def push(self, key: str, value: T) -> Either[QueueError, None]:
    r = await self._write_req('POST', 'push', data=self.dump(value), params=dict(key=key))
    return r.fmap(lambda _: None).mapl(QueueError)


from typing import Any
from haskellian import Left
from pydantic import TypeAdapter, ValidationError
from pipeteer.queues import QueueError, ReadError

def urljoin(part1: str, part2: str):
  """What `urllib.parse.urljoin` ought to be."""
  return part1.rstrip('/') + '/' + part2.lstrip('/')

ErrType = TypeAdapter(ReadError)

def validate_left(raw_json: bytes, status: int) -> Left[QueueError, Any]:
  try:
    return Left(ErrType.validate_json(raw_json))
  except ValidationError:
    return Left(QueueError(f'Unexpected status code: {status}. Content: "{raw_json.decode()}"'))
  



from typing import Protocol, Literal, Mapping

class Response(Protocol):
  status_code: int
  content: bytes

class Request(Protocol):
  async def __call__(
    self, method: Literal['GET', 'POST', 'DELETE'], url: str, /, *,
    data: bytes | str | None = None, params: Mapping[str, str] = {}
  ) -> Response: ...

async def request(
  method: Literal['GET', 'POST', 'DELETE'], url: str, /, *,
  data: bytes | str | None = None, params: Mapping[str, str] = {}, **kwargs
) -> Response:
  import httpx
  async with httpx.AsyncClient() as client:
    return await client.request(method, url, data=data, params=params, **kwargs) # type: ignore
  
def bound_request(*, headers: Mapping[str, str], **kwargs) -> Request:
  async def _bound(
    method: Literal['GET', 'POST', 'DELETE'], url: str, /, *,
    data: bytes | str | None = None, params: Mapping[str, str] = {}
  ):
    return await request(method, url, data=data, params=params, headers=headers, **kwargs)
  
  return _bound


from typing import TypeVar, Generic, Mapping, Union, Any
from abc import abstractmethod
from pipeteer.queues import WriteQueue, ops
from pipeteer import Pipeline, GetQueue, trees

A = TypeVar('A')
B = TypeVar('B')
Q = TypeVar('Q', bound=Mapping)
P = TypeVar('P')
Ps = TypeVar('Ps', bound=Mapping[str, Pipeline])
T = TypeVar('T')

class Workflow(Pipeline[A, B, Q, P, T], Generic[A, B, Q, P, T, Ps]):
  """State-machine-like composition of pipelines"""

  def __init__(self, pipelines: Ps, *, Tin: type[A] | None = None, Tout: type[B] | None = None):
    self.Tin = Tin or Union[*(pipe.Tin for pipe in pipelines.values())] # type: ignore
    self.Tout = Tout
    self.pipelines = pipelines

  def __repr__(self):
    from .reprs import str_types, indent
    inner = '\n'.join(f'{id}: {repr(pipe)},' for id, pipe in self.pipelines.items())
    return '\n'.join([
      f'Workflow[{str_types(self)}](',
        indent(inner, 2),
      ')',
    ])
  
  @staticmethod
  def dict(pipelines: Mapping[str, Pipeline[Any, Any, Any, P, T]]) -> 'DictWorkflow[A, B, Q, P, T]':
    return DictWorkflow(pipelines)
  
  def push_queue(self, get_queue: GetQueue, params: P, *, prefix: tuple[str|int, ...] = ()) -> WriteQueue[A]:
    outputs = [(pipe.push_queue(get_queue, params, prefix=prefix + (id,)), pipe.Tin) for id, pipe in self.pipelines.items()]
    return ops.prejoin(outputs)
  
  def connect(self, Qout: WriteQueue[B], get_queue: GetQueue, params: P, *, prefix: tuple[str|int, ...] = ()) -> Q:
    outputs = [(pipe.push_queue(get_queue, params, prefix=prefix + (id,)), pipe.Tin) for id, pipe in self.pipelines.items()]
    Qpush = ops.prejoin(outputs, fallback=Qout)
    queues = {id: pipe.connect(Qpush, get_queue, params, prefix=prefix + (id,)) for id, pipe in self.pipelines.items()}
    return queues # type: ignore
  
  def tree(self) -> trees.Tree[Pipeline]:
    return {id: pipe.tree() for id, pipe in self.pipelines.items()}

  @abstractmethod
  def run(self, queues: Q, params: P, /) -> T:
    ...


class DictWorkflow(Workflow[A, B, Mapping[str, Q], P, dict[str, T], Mapping[str, Pipeline]], Generic[A, B, Q, P, T]):
  def run(self, queues: Mapping[str, Q], params: P, /) -> dict[str, T]: # type: ignore
    return {id: pipe.run(queues[id], params) for id, pipe in self.pipelines.items()}


from typing import TypeVar, Generic, AsyncIterable, Callable, Sequence, Literal
import asyncio
from datetime import timedelta
import json
from pydantic import TypeAdapter
from haskellian import Either, Left, Right, asyn_iter as AI
from pipeteer.queues import ReadQueue, ReadError, QueueError, InexistentItem
from .util import urljoin, validate_left
from .request import request, Request

T = TypeVar('T')

KeysType = TypeAdapter(Sequence[Either[QueueError, str]])

def validate_seq(raw_json: bytes) -> Sequence[Either[QueueError, str]]:
  try:
    return KeysType.validate_json(raw_json)
  except Exception as e:
    return [Left(QueueError(e))]

  
class ReadClient(ReadQueue[T], Generic[T]):
    
  def __init__(
    self, url: str, *,
    parse: Callable[[bytes], Either[QueueError, T]] = Right,
    polling_interval = timedelta(seconds=15),
    request: Request = request
  ):
    self.read_url = url
    self.parse = parse
    self.polling_interval = polling_interval
    self.request = request

  async def _read_req(
    self, method: Literal['GET', 'POST', 'DELETE'], path: str, *,
    data: bytes | str | None = None, params: dict = {}
  ):
    try:
      r = await self.request(method, urljoin(self.read_url, path), data=data, params=params)
      return Right(r.content) if r.status_code == 200 else validate_left(r.content, r.status_code)
    except Exception as e:
      return Left(QueueError(e))

  @classmethod
  def validated(cls, Type: type[T], url: str, *, polling_interval = timedelta(seconds=15), request: Request = request) -> 'ReadClient[T]':
    from pydantic import TypeAdapter
    Adapter = TypeAdapter(Type)
    def parse(x):
      try:
        return Right(Adapter.validate_json(x))
      except Exception as e:
        return Left(QueueError(str(e)))
    return cls(url, parse=parse, polling_interval=polling_interval, request=request)


  async def _read(self, id: str | None) -> Either[ReadError, tuple[str, T]]: # type: ignore
    
    if id is None: # read/any -> returns an arbitrary id
      r = await self._read_req('GET', 'read/any')
      if r.tag == 'left':
        if r.value.reason == 'inexistent-item':
          await asyncio.sleep(self.polling_interval.total_seconds())
          return await self._read(None)

      id: str = json.loads(r.unsafe())

    data = await self._read_req('GET', 'read', params=dict(id=id))
    return data.bind(self.parse).fmap(lambda val: (id, val))
  
  async def pop(self, id: str) -> Either[QueueError | InexistentItem, None]:
    return (await self._read_req('GET', 'pop', params=dict(id=id))).fmap(lambda _: None)
    
  async def _items(self) -> AsyncIterable[Either[QueueError, tuple[str, T]]]:
    keys = await self.keys().sync()
    for key in keys:
      if key.tag == 'left':
        yield Left(key.value)
      else:
        yield (await self._read(key.value)).mapl(QueueError)

  @AI.lift
  async def keys(self) -> AsyncIterable[Either[QueueError, str]]:
    r = await self._read_req('GET', 'keys')
    if r.tag == 'left':
      yield r.mapl(QueueError)
    else:
      for key in validate_seq(r.value):
        yield key


from .read import ReadQueue
from .write import WriteQueue
from .queue import Queue
from .errors import QueueError, InexistentItem, ReadError
from .impl import SimpleQueue, QueueKV
from . import ops

__all__ = [
  'ReadQueue', 'WriteQueue', 'Queue',
  'QueueError', 'InexistentItem', 'ReadError',
  'SimpleQueue', 'QueueKV',
  'ops',
]


from typing_extensions import AsyncIterable, Generic, TypeVar
import asyncio
from datetime import timedelta
import random
from haskellian import Either, Left, IsLeft, Right
from kv import KV, InexistentItem as KVInexistentItem
from pipeteer.queues import Queue, QueueError, ReadError, InexistentItem

A = TypeVar('A')
B = TypeVar('B')

class QueueKV(Queue[A], Generic[A]):
  """Poll-based queue backed by a `KV` object"""
  def __init__(
    self, kv: KV[A],
    poll_interval: timedelta = timedelta(seconds=5),
  ):
    self._kv = kv
    self.poll_interval = poll_interval

  def __repr__(self):
    return f'QueueKV({type(self._kv).__name__})'

  @staticmethod
  def of(conn_str: str, type: type[A] | None = None) -> 'QueueKV[A]':
    """Create a `Queue` with a `KV` from a connection string"""
    return QueueKV(KV.of(conn_str, type))
  
  @staticmethod
  def sqlite(Type: type[B], path: str, table: str = 'queue') -> 'QueueKV[B]':
    from kv import SQLiteKV
    kv = SQLiteKV.at(path, Type, table=table)
    return QueueKV(kv)
  
  async def push(self, key: str, value: A):
    return (await self._kv.insert(key, value)).mapl(QueueError)
  
  async def _read(self, id: str | None = None) -> Either[ReadError, tuple[str, A]]:
    if id is None:
      keys = await self._kv.keys().sync()
      if keys:
        key = random.choice(keys).unsafe()
      else:
        await asyncio.sleep(self.poll_interval.total_seconds())
        return await self._read(id)
    else:
      key = id
    
    try:
      value = (await self._kv.read(key)).unsafe()
      return Right((key, value))
    except IsLeft as e:
      match e.value:
        case KVInexistentItem():
          return Left(InexistentItem(key))
        case err:
          return Left(QueueError(err))
        
  async def pop(self, id: str) -> Either[QueueError | InexistentItem, None]:
    match await self._kv.delete(id):
      case Left(KVInexistentItem()):
        return Left(InexistentItem(id))
      case Left(err):
        return Left(QueueError(err))
      case _:
        return Right(None)
    
  def _items(self) -> AsyncIterable[Either[QueueError, tuple[str, A]]]:
    return self._kv.items().map(lambda e: e.mapl(QueueError))
  


from typing_extensions import TypeVar, Generic, Callable, Awaitable, AsyncIterable
from dataclasses import dataclass
from haskellian import Either, Right
from pipeteer.queues import ReadQueue, QueueError, ReadError

A = TypeVar('A')
B = TypeVar('B')

@dataclass
class immutable(ReadQueue[A], Generic[A]):
  """An immutable view of a `ReadQueue`"""
  queue: ReadQueue[A]

  async def _read(self, id: str | None):
    return await self.queue._read(id)
  
  async def pop(self, id: str):
    return Right(None)
  
  def _items(self):
    return self.queue._items()
  
  def __getattr__(self, name: str):
    return getattr(self.queue, name)
  
class map(ReadQueue[B], Generic[A, B]):
  
  def __init__(self, q: ReadQueue[A], f: Callable[[tuple[str, A]], Awaitable[tuple[str, B]]]):
    self._queue = q
    self._mapper = f

  def __repr__(self):
    return f'mapped({repr(self._queue)})'

  async def _items(self) -> AsyncIterable[Either[QueueError, tuple[str, B]]]:
    async for e in self._queue._items():
      if e.tag == 'left':
        yield e # type: ignore
      else:
        yield Right(await self._mapper(e.value))

  async def _read(self, id: str | None) -> Either[ReadError, tuple[str, B]]:
    match await self._queue._read(id):
      case Right((k, v)):
        return Right(await self._mapper((k, v)))
      case err:
        return err # type: ignore
      
  async def pop(self, id: str):
    return await self._queue.pop(id)
      
class filter(ReadQueue[A], Generic[A]):

  def __init__(self, queue: ReadQueue[A], pred: Callable[[str, A], bool]):
    self._pred = pred
    self._queue = queue

  def __repr__(self):
    return f'filtered({repr(self._queue)})'

  async def _point_read(self, id: str) -> Either[ReadError, tuple[str, A]]:
    e = await self._queue.read(id)
    if e.tag == 'left':
      return e # type: ignore
    item = e.value
    return Right((id, item))

  async def _read_any(self) -> Either[QueueError, tuple[str, A]]:
    async for e in self._items():
      if e.tag == 'left':
        return e
      id, item = e.value
      return Right((id, item))
    return await self._read_any()

  async def _read(self, id: str | None) -> Either[ReadError, tuple[str, A]]:
    return await (self._read_any() if id is None else self._point_read(id))
  
  async def pop(self, id: str):
    return await self._queue.pop(id)
  
  async def _items(self) -> AsyncIterable[Either[QueueError, tuple[str, A]]]:
    async for e in self._queue._items():
      if e.tag == 'left':
        yield e
      elif self._pred(*e.value):
        yield e


from .pipeline import Pipeline, GetQueue
from .task import Task
from .wrapped import Wrapped
from .workflow import Workflow, DictWorkflow
from .sequential import Sequential

__all__ = [
  'Pipeline', 'GetQueue',
  'Task', 'Wrapped', 'Sequential',
  'Workflow', 'DictWorkflow',
]


from typing import TypeVar, Mapping, Sequence, Iterable, Callable
from haskellian import iter as I


A = TypeVar('A')
B = TypeVar('B')
T = TypeVar('T')

Tree = T | Mapping[str, 'Tree[T]'] | Sequence['Tree[T]']

@I.lift
def flatten(tree: Tree[T], prefix: tuple[str|int, ...] = ()) -> Iterable[tuple[Sequence[str|int], T]]:
  if isinstance(tree, Mapping):
    for k, v in tree.items():
      yield from flatten(v, prefix + (k,))
  elif isinstance(tree, Sequence):
    for i, v in enumerate(tree):
      yield from flatten(v, prefix + (i,))
  else:
    yield prefix, tree


def map(tree: Tree[A], f: Callable[[A], B]) -> Tree[B]:
  if isinstance(tree, Mapping):
    return {k: map(v, f) for k, v in tree.items()}
  elif isinstance(tree, Sequence):
    return [map(v, f) for v in tree]
  else:
    return f(tree)
  
def path_map(tree: Tree[A], f: Callable[[Sequence[str|int], A], B], prefix: tuple[str|int, ...] = ()) -> Tree[B]:
  if isinstance(tree, Mapping):
    return {k: path_map(v, f, prefix + (k,)) for k, v in tree.items()}
  elif isinstance(tree, Sequence):
    return [path_map(v, f, prefix + (i,)) for i, v in enumerate(tree)]
  else:
    return f(prefix, tree)


from typing_extensions import Literal, Any
from dataclasses import dataclass

@dataclass(eq=False)
class QueueError(BaseException):
  detail: Any | None = None
  reason: Literal['queue-error'] = 'queue-error'

  def __str__(self) -> str:
    return self.__repr__()

@dataclass(eq=False)
class InexistentItem(BaseException):
  key: str | None = None
  detail: Any | None = None
  reason: Literal['inexistent-item'] = 'inexistent-item'

  def __str__(self) -> str:
    return self.__repr__()
  
ReadError = QueueError | InexistentItem


from typing import TypeVar, Sequence, Callable, overload
from haskellian import Either, Left
from pipeteer.queues import ReadQueue, ReadError, InexistentItem
from fastapi import FastAPI, Response
from pydantic import TypeAdapter

T = TypeVar('T')
U = TypeVar('U')

@overload
def read_api(
  queue: ReadQueue[T], *,
  dump: Callable[[T], bytes|str] = lambda x: x # type: ignore
) -> FastAPI:
  ...
@overload
def read_api(
  queue: ReadQueue[T], *, Type: type[T]
) -> FastAPI:
  ...
def read_api(queue, *, dump = lambda x: x, Type = None): # type: ignore
  if Type is not None:
    from pydantic import TypeAdapter
    dump = TypeAdapter(Type).dump_json
  
  return _read_api(queue, dump=dump)

def _read_api(
  queue: ReadQueue[T], *,
  dump: Callable[[T], bytes|str] = lambda x: x # type: ignore
) -> FastAPI:
  """
  - `timeout`: timeout for read operations.
  """

  from .util import status, with_status
  app = FastAPI(generate_unique_id_function=lambda route: route.name)

  @app.get('/read/any')
  @with_status
  async def read_any():
    async for e in queue.keys():
      if e.tag == 'right':
        return e
    
    return Left(InexistentItem(detail='No keys found'))

  @app.get('/read')
  async def read(id: str):
    e = await queue._read(id)
    if e.tag == 'right':
      _, v = e.value
      return Response(dump(v))
    else:
      content = TypeAdapter(ReadError).dump_json(e.value)
      return Response(content, status_code=status(e))
    
  @app.get('/pop')
  @with_status
  async def pop(id: str):
    return await queue.pop(id)
  
  @app.get('/keys')
  async def keys() -> Sequence[Either[ReadError, str]]:
    return await queue.keys().sync()

  return app


from .write import prejoin, premerge, tee, premap, prefilter
from .read import immutable, map, filter

__all__ = [
  'prejoin', 'premerge', 'tee', 'premap', 'prefilter',
  'immutable', 'map', 'filter',
]


from typing_extensions import Generic, TypeVar, Callable, Awaitable, AsyncIterable, TypeGuard, overload, Any
from haskellian import promise as P, Either, Right, either as E
from abc import ABC, abstractmethod
from .errors import QueueError
from . import ops

A = TypeVar('A', contravariant=True)
B = TypeVar('B', contravariant=True)

class WriteQueue(ABC, Generic[A]):
  """A write-only queue"""
  
  async def iterate(self, items: AsyncIterable[tuple[str, A]]):
    """Push all `items`"""
    async for key, value in items:
      await self.push(key, value)

  @abstractmethod
  async def push(self, key: str, value: A) -> Either[QueueError, None]:
    ...

  def pusher(self, key: str) -> Callable[[A], Awaitable[Either[QueueError, None]]]:
    """Partially applied `push`"""
    return lambda value: self.push(key, value)
  
  @overload
  def prefilter(self, pred: Callable[[A], TypeGuard[B]]) -> 'WriteQueue[B]': ...
  @overload
  def prefilter(self, pred: Callable[[A], bool]) -> 'WriteQueue[A]': ...
  def prefilter(self, pred): # type: ignore
    return ops.prefilter(self, lambda kv: P.of(pred(kv[1])))
  
  def premap(self, f: Callable[[B], A]) -> 'WriteQueue[B]':
    return ops.premap(self, lambda kv: P.of(Right((kv[0], f(kv[1])))))
  
  def premap_kv(self, f: Callable[[str, B], A]) -> 'WriteQueue[B]':
    """Map but `f` receives both key and value"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], f(*kv)
    return ops.premap(self, mapper)
  
  def premap_k(self, f: Callable[[str], A]) -> 'WriteQueue':
    """Map but `f` receives the key"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], f(kv[1])
    return ops.premap(self, mapper)
  
  def premap_kvt(self, f: Callable[[tuple[str, B]], A]) -> 'WriteQueue[B]':
    """Map but `f` receives both key and value as a tuple"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], f(kv)
    return ops.premap(self, mapper)
  
  def apremap(self, f: Callable[[B], Awaitable[A]]) -> 'WriteQueue[B]':
    """Map but `f` is asynchronous"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], await f(kv[1])
    return ops.premap(self, mapper)
  
  def safe_apremap(self, f: Callable[[B], Awaitable[Either[Any, A]]]) -> 'WriteQueue[B]':
    """Map but `f` is asynchronous and can fail (ie. return `Left`)"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], (await f(kv[1])).unsafe()
    return ops.premap(self, mapper)
  
  def apremap_kv(self, f: Callable[[str, B], Awaitable[A]]) -> 'WriteQueue[B]':
    """Map but `f` is asynchronous and receives both key and value"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], await f(*kv)
    return ops.premap(self, mapper)
  
  def apremap_k(self, f: Callable[[str], Awaitable[A]]) -> 'WriteQueue':
    """Map but `f` is asynchronous and receives the key"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], await f(kv[0])
    return ops.premap(self, mapper)
  
  def apremap_kvt(self, f: Callable[[tuple[str, B]], Awaitable[A]]) -> 'WriteQueue[B]':
    """Map but `f` is asynchronous and receives both key and value as a tuple"""
    @E.do()
    async def mapper(kv: tuple[str, B]):
      return kv[0], await f(kv)
    return ops.premap(self, mapper)




from typing import ParamSpec, TypeVar, Awaitable, Callable, Any
from functools import wraps
import inspect
from fastapi import Response, status as st
from haskellian import Either, kwargs as kw
from pipeteer.queues import ReadError

T = TypeVar('T')
Ps = ParamSpec('Ps')

def status(e: Either[ReadError, T]):
  if e.tag == 'right':
    return st.HTTP_200_OK
  elif e.value.reason == 'queue-error':
    return st.HTTP_500_INTERNAL_SERVER_ERROR
  else:
    return st.HTTP_404_NOT_FOUND

def with_status(func: Callable[Ps, Awaitable[Either[ReadError, T]]]):
  @wraps(func)
  async def wrapper(response: Response, *args: Ps.args, **kwargs: Ps.kwargs) -> ReadError | T:
    e = await func(*args, **kwargs)
    response.status_code = status(e)
    return e.value
  wrapper.__signature__ = kw.add_kw(inspect.signature(wrapper), 'response', Response) # type: ignore
  return wrapper


from typing import Any, NamedTuple, TypeVar, Generic
from pipeteer import Pipeline, http, trees
from pipeteer.http.client.request import Request, request

A = TypeVar('A')
B = TypeVar('B')

class PipelineClients(NamedTuple, Generic[A, B]):
  Qin: http.WriteClient[A]
  Qout: http.ReadClient[B]
  Qs: trees.Tree[http.ReadClient]

def input_client(pipe: Pipeline, base_url: str, *, request: Request = request):
  return http.WriteClient.validated(pipe.Tin, base_url + '/input/write', request=request)

def output_client(pipe: Pipeline, base_url: str, *, request: Request = request):
  return http.ReadClient.validated(pipe.Tout or Any, base_url + '/output/read', request=request)

def queue_clients(pipe: Pipeline, base_url: str, *, request: Request = request):
  base_url = base_url.rstrip('/') + '/'
  get_client = lambda path, _: http.ReadClient.validated(Any, base_url + '/'.join(path) + '/read', request=request)
  return trees.path_map(pipe.tree(), get_client)

def clients(pipe: Pipeline[A, B, Any, Any, Any], base_url: str, *, request: Request = request):
  """Clients for a pipeline's queues, as exposed by `http.mount`"""
  Qin = input_client(pipe, base_url, request=request)
  Qout = output_client(pipe, base_url, request=request)
  Qs = queue_clients(pipe, base_url, request=request)
  return PipelineClients[A, B](Qin, Qout, Qs)


from typing import Any, Generic, Sequence, TypeVar
from pipeteer import Pipeline, trees, GetQueue, WriteQueue

A = TypeVar('A')
B = TypeVar('B')
Q = TypeVar('Q')
P = TypeVar('P')
T = TypeVar('T')

class Sequential(Pipeline[A, B, Sequence[Q], Sequence[P], Sequence[T]], Generic[A, B, Q, P, T]):
  """Sequential composition of pipelines, with a few caveats
  - The tree of queues is a list of the pipeline queues in the given order
  - The expected parameters are a list of each pipeline's parameters. There must be exactly one per pipeline
  """
  def __init__(self, pipelines: Sequence[Pipeline[Any, Any, Q, P, T]]):
    assert len(pipelines) > 0, "Sequential pipelines must have at least one pipeline"
    self.pipelines = pipelines
    self.Tin = pipelines[0].Tin
    self.Tout = pipelines[-1].Tout

  def tree(self) -> trees.Tree[Pipeline]:
    return [p.tree() for p in self.pipelines]
  
  def connect(self, Qout: WriteQueue[B], get_queue: GetQueue, params: Sequence[P], *, prefix: tuple[str|int, ...] = ()):
    queues = []
    output = Qout
    params = params or [None] * len(self.pipelines) # type: ignore
    for i, (pipe, ps) in reversed(list(enumerate(zip(self.pipelines, params)))):
      i_prefix = prefix + (i,)
      queues.append(pipe.connect(output, get_queue, ps, prefix=i_prefix))
      output = pipe.push_queue(get_queue, ps, prefix=i_prefix)
    return list(reversed(queues))

  def push_queue(self, get_queue: GetQueue, params: Sequence[P], *, prefix: tuple[str|int, ...] = ()) -> WriteQueue[A]:
    return self.pipelines[0].push_queue(get_queue, params[0], prefix=prefix + (0,))
  
  def run(self, queues: Sequence[Q], params: Sequence[P] = None): # type: ignore
    params = params or [None] * len(self.pipelines) # type: ignore
    return [pipe.run(Qs, ps) for pipe, ps, Qs in zip(self.pipelines, params, queues)]
  


from typing_extensions import Generic, TypeVar, Callable, Awaitable, overload, AsyncIterable, TypeGuard, Sequence
from abc import ABC, abstractmethod
from haskellian import AsyncIter, promise as P, Either, Right, iter as I, either as E
from .errors import QueueError, ReadError
from . import ops

A = TypeVar('A', covariant=True)
B = TypeVar('B', covariant=True)

class ReadQueue(ABC, Generic[A]):
  """A read/pop-only, point-readable queue"""

  @abstractmethod
  async def _read(self, id: str | None) -> Either[ReadError, tuple[str, A]]:
    ...

  @abstractmethod
  async def pop(self, id: str) -> Either[ReadError, None]:
    ...

  @overload
  async def read(self) -> Either[QueueError, tuple[str, A]]:
    ...
  @overload
  async def read(self, id: str) -> Either[ReadError, A]:
    ...
  
  @E.do[ReadError]()
  async def read(self, id: str | None = None) -> A | tuple[str, A]:
    k, v = (await self._read(id)).unsafe()
    return (k, v) if id is None else v
      
  @abstractmethod
  def _items(self) -> AsyncIterable[Either[QueueError, tuple[str, A]]]:
    ...

  def items(self) -> AsyncIter[Either[QueueError, tuple[str, A]]]:
    return AsyncIter(self._items())
  
  def keys(self) -> AsyncIter[Either[ReadError, str]]:
    return AsyncIter(e | I.fst async for e in self._items())
  
  def values(self) -> AsyncIter[Either[ReadError, A]]:
    return AsyncIter(e | I.snd async for e in self._items())
  
  def map(self, f: Callable[[A], B]) -> 'ReadQueue[B]':
    """Maps `f` over self. Returns a new queue, but `self` is still mutated when popping from the new queue"""
    return ops.map(self, lambda kv: P.of((kv[0], f(kv[1]))))
  
  def map_kv(self, f: Callable[[str, A], B]) -> 'ReadQueue[B]':
    """Map but `f` receives both key and value"""
    return ops.map(self, lambda kv: P.of((kv[0], f(*kv))))
  
  def map_k(self, f: Callable[[str], B]) -> 'ReadQueue[B]':
    """Map but `f` receives the key"""
    return ops.map(self, lambda kv: P.of((kv[0], f(kv[0]))))
  
  def map_kvt(self, f: Callable[[tuple[str, A]], B]) -> 'ReadQueue[B]':
    """Map but `f` receives both key and value as a tuple"""
    return ops.map(self, lambda kv: P.of((kv[0], f(kv))))
  
  def amap(self, f: Callable[[A], Awaitable[B]]) -> 'ReadQueue[B]':
    """Map but `f` is asynchronous"""
    async def mapper(kv: tuple[str, A]):
      return kv[0], await f(kv[1])
    return ops.map(self, mapper)
  
  def amap_kv(self, f: Callable[[str, A], Awaitable[B]]) -> 'ReadQueue[B]':
    """Map but `f` is asynchronous and receives both key and value"""
    async def mapper(kv: tuple[str, A]):
      return kv[0], await f(*kv)
    return ops.map(self, mapper)
  
  def amap_k(self, f: Callable[[str], Awaitable[B]]) -> 'ReadQueue[B]':
    """Map but `f` is asynchronous and receives the key"""
    async def mapper(kv: tuple[str, A]):
      return kv[0], await f(kv[0])
    return ops.map(self, mapper)
  
  def amap_kvt(self, f: Callable[[tuple[str, A]], Awaitable[B]]) -> 'ReadQueue[B]':
    """Map but `f` is asynchronous and receives both key and value as a tuple"""
    async def mapper(kv: tuple[str, A]):
      return kv[0], await f(kv)
    return ops.map(self, mapper)

  @overload
  def filter(self, pred: Callable[[A], TypeGuard[B]]) -> 'ReadQueue[B]': ...
  @overload
  def filter(self, pred: Callable[[A], bool]) -> 'ReadQueue[A]': ...
  def filter(self, pred): # type: ignore
    return ops.filter(self, lambda _, v: pred(v))
  
  @overload
  def filter_kv(self, pred: Callable[[str, A], TypeGuard[B]]) -> 'ReadQueue[B]': ...
  @overload
  def filter_kv(self, pred: Callable[[str, A], bool]) -> 'ReadQueue[A]': ...
  def filter_kv(self, pred): # type: ignore
    return ops.filter(self, pred)
  
  def partition(self, pred: Callable[[A], bool]) -> 'tuple[ReadQueue[A], ReadQueue[A]]':
    """Returns `self.filter(pred), self.filter(!pred)`"""
    return self.filter(pred), self.filter(lambda x: not pred(x))
  
  def partition_kv(self, pred: Callable[[str, A], bool]) -> 'tuple[ReadQueue[A], ReadQueue[A]]':
    """Returns `self.filter_kv(pred), self.filter_kv(!pred)`"""
    return self.filter_kv(pred), self.filter_kv(lambda *x: not pred(*x))
  
  




from typing import TypeVar, Callable, overload
from pipeteer.queues import WriteQueue, QueueError
from fastapi import FastAPI, Request
from haskellian import either as E, Either, Right, Left

T = TypeVar('T')

@overload
def write_api(
  queue: WriteQueue[T], *,
  parse: Callable[[bytes], Either[QueueError, T]] = Right
) -> FastAPI:
  ...
@overload
def write_api(
  queue: WriteQueue[T], *, Type: type[T]
) -> FastAPI:
  ...
def write_api(queue, *, parse = Right, Type = None): # type: ignore
  if Type is not None:
    from pydantic import TypeAdapter
    Adapter = TypeAdapter(Type)
    def parse(x):
      try:
        return Right(Adapter.validate_json(x))
      except Exception as e:
        return Left(QueueError(str(e)))
  
  return _write_api(queue, parse=parse)
  
def _write_api(
  queue: WriteQueue[T], *,
  parse: Callable[[bytes], Either[QueueError, T]] = Right
) -> FastAPI:

  from .util import with_status
  app = FastAPI(generate_unique_id_function=lambda route: route.name)

  @app.post('/push')
  @with_status
  @E.do[QueueError]()
  async def push(key: str, req: Request):
    body = await req.body()
    val = parse(body).unsafe()
    return (await queue.push(key, val)).unsafe()

  return app


from typing import TypeVar, Generic, TypedDict, Any, Callable, Awaitable
from types import UnionType
from dataclasses import dataclass
from abc import abstractmethod
from haskellian import funcs as F, promise
from pipeteer.queues import WriteQueue, Queue, ops
from pipeteer import Pipeline, GetQueue, trees

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
Q = TypeVar('Q')
Q2 = TypeVar('Q2')
P = TypeVar('P')
P2 = TypeVar('P2')
T = TypeVar('T')
T2 = TypeVar('T2')
S1 = TypeVar('S1')
S2 = TypeVar('S2')
S3 = TypeVar('S3')
S4 = TypeVar('S4')

class WrappedQueues(TypedDict, Generic[S1, Q]):
  wrapper: Queue[S1]
  wrapped: Q

@dataclass
class Wrapped(Pipeline[S1, S2, WrappedQueues[S1, Q], P, T], Generic[S1, S2, A, B, Q, P, T]):
  """Wrap a pipeline with pre- and post-processing functions. Both wrapping and wrapped values are stored in a queue"""

  Queues = WrappedQueues

  def __init__(
    self, Tin: type[S1], pipeline: Pipeline[A, B, Q, P, Any], *,
    Tout: type[S2] | UnionType | None = None
  ):
    self.Tin = Tin
    self.Tout = Tout
    self.pipeline = pipeline

  def __repr__(self):
    from .reprs import str_types, indent
    inner = repr(self.pipeline)
    if len(inner) > 60:
      inner = '\n' + indent(inner) + '\n'
    return f'Wrapped[{str_types(self)}]({inner})'

  @staticmethod
  def of(
    Tin: type[S3], pipeline: Pipeline[C, D, Q2, P2, T2],
    pre: Callable[[S3], Awaitable[C] | C], post: Callable[[S3, D], Awaitable[S4] | S4], *,
    Tout: type[S4] | UnionType | None = None
  ) -> 'Wrapped[S3, S4, C, D, Q2, P2, T2]':
    return FuncWrapped(Tin, pipeline, pre, post, Tout=Tout)

  @abstractmethod
  def pre(self, inp: S1, /) -> Awaitable[A] | A:
    ...
  
  @abstractmethod
  def post(self, inp: S1, out: B, /) -> Awaitable[S2] | S2:
    ...

  def push_queue(self, get_queue: GetQueue, params: P, *, prefix: tuple[str|int, ...] = ()) -> WriteQueue[S1]:
    Qwrapper = get_queue(prefix + ('wrapper',), self.Tin)
    Qpush = self.pipeline.push_queue(get_queue, params, prefix=prefix + ('wrapped',))
    return ops.tee(Qwrapper, Qpush.apremap(F.flow(self.pre, promise.wait)))
  
  def connect(self, Qout: WriteQueue[S2], get_queue: GetQueue, params: P, *, prefix: tuple[str|int, ...] = ()) -> WrappedQueues[S1, Q]:
    Qwrapper = get_queue(prefix + ('wrapper',), self.Tin)
    wrapped_Qout = ops.premerge(Qwrapper, Qout, self.post)
    wrapped = self.pipeline.connect(wrapped_Qout, get_queue, params, prefix=prefix + ('wrapped',))
    return WrappedQueues(wrapper=Qwrapper, wrapped=wrapped)
  
  def tree(self) -> trees.Tree[Pipeline]:
    return { 'wrapper': self, 'wrapped': self.pipeline.tree() }
  
  def run(self, queues: WrappedQueues[S1, Q], params: P, /) -> T:
    return self.pipeline.run(queues['wrapped'], params)
  

class FuncWrapped(Wrapped[S1, S2, A, B, Q, P, T], Generic[S1, S2, A, B, Q, P, T]):
  def __init__(
    self, Tin: type[S1], pipeline: Pipeline[A, B, Q, P, Any], pre: Callable[[S1], Awaitable[A] | A], post: Callable[[S1, B], Awaitable[S2] | S2], *,
    Tout: type[S2] | UnionType | None = None
  ):
    super().__init__(Tin, pipeline, Tout=Tout)
    self._pre = pre
    self._post = post

  def pre(self, inp: S1, /) -> Awaitable[A] | A:
    return self._pre(inp)
  
  def post(self, inp: S1, out: B, /) -> Awaitable[S2] | S2:
    return self._post(inp, out)


from typing_extensions import Generic, TypeVar, AsyncIterable, Never
from collections import OrderedDict
from haskellian import ManagedPromise, Either, Left, Right
from pipeteer.queues import Queue, InexistentItem

A = TypeVar('A')
B = TypeVar('B')

class SimpleQueue(Queue[A], Generic[A]):
  """Reactive, in-memory implementation backed by an `OrderedDict`"""

  def __init__(self, name: str | None = None):
    self.xs: OrderedDict[str, A] = OrderedDict()
    self._next = ManagedPromise()
    self.name = name

  def __repr__(self):
    return f'SimpleQueue({self.name or ""})'

  def __len__(self):
    return len(self.xs) # type: ignore

  async def _read(self, id: str | None = None) -> Either[InexistentItem, tuple[str, A]]:
    if id is None:
      if len(self.xs) == 0:
        await self._next
        self._next = ManagedPromise()
        return await self._read(id)
      else:
        return Right(next(iter(self.xs.items())))
    elif id in self.xs:
      v = self.xs[id]
      return Right((id, v))
    else:
      return Left(InexistentItem(id))
    
  async def pop(self, id: str):
    if id in self.xs:
      del self.xs[id]
      return Right(None)
    else:
      return Left(InexistentItem(id))
    
  async def push(self, key: str, value: A) -> Right[Never, None]:
    self.xs[key] = value
    if not self._next.resolved:
      self._next.resolve()
    return Right(None)
  
  async def _items(self) -> AsyncIterable[Right[Never, tuple[str, A]]]:
    for x in self.xs.items():
      yield Right(x)
  


from typing import TypeVar, Callable, overload
from pipeteer.queues import Queue, QueueError
from fastapi import FastAPI
from pydantic import TypeAdapter
from haskellian import Right, Left, Either
from .read import read_api
from .write import write_api

T = TypeVar('T')
U = TypeVar('U')


def _api(
  queue: Queue[T], *,
  parse: Callable[[bytes], Either[QueueError, T]] = Right,
  dump: Callable[[T], bytes|str] = lambda x: x, # type: ignore
) -> FastAPI:
  app = FastAPI(generate_unique_id_function=lambda route: route.name)
  app.mount('/read', read_api(queue, dump=dump))
  app.mount('/write', write_api(queue, parse=parse))

  return app
  
@overload
def queue_api(queue: Queue[T], *, Type: type[T]) -> FastAPI: ...
@overload
def queue_api(
  queue: Queue[T], *,
  parse: Callable[[bytes], Either[QueueError, T]] = Right,
  dump: Callable[[T], bytes] = lambda x: x # type: ignore
) -> FastAPI: ...

def queue_api(queue, Type=None, dump=lambda x: x, parse=Right): # type: ignore
  if Type is not None:
    Adapter = TypeAdapter(Type)
    dump = Adapter.dump_json
    def parse(x):
      try:
        return Right(Adapter.validate_json(x))
      except Exception as e:
        return Left(QueueError(str(e)))
  return _api(queue, dump=dump, parse=parse)



from typing import TypeVar, Generic, Protocol, Mapping, Sequence
from types import UnionType
from dataclasses import dataclass
from abc import ABC, abstractmethod
from pipeteer import WriteQueue, ReadQueue, Queue, trees

A = TypeVar('A', covariant=True)
B = TypeVar('B', covariant=True)
C = TypeVar('C')
D = TypeVar('D', covariant=True)
Q = TypeVar('Q')
P = TypeVar('P')
T = TypeVar('T')
S1 = TypeVar('S1')
S2 = TypeVar('S2')

class GetQueue(Protocol):
  """Lazy queue accessor. Must return *the same object* given the same parameters."""
  def __call__(self, path: Sequence[str|int], type: type[T], /) -> Queue[T]:
    ...

@dataclass
class Pipeline(ABC, Generic[A, B, Q, P, T]):
  """Base class for all pipelines"""
  Tin: type[A]
  Tout: type[B] | UnionType | None

  def __repr__(self):
    from .reprs import str_types
    return f'Pipeline[{str_types(self)}]'

  @abstractmethod
  def push_queue(self, get_queue: GetQueue, params: P, /, *, prefix: tuple[str|int, ...] = ()) -> WriteQueue[A]:
    """Queue to push tasks into the pipeline"""

  @abstractmethod
  def connect(self, Qout: WriteQueue[B], get_queue: GetQueue, params: P, /, *, prefix: tuple[str|int, ...] = ()) -> Q:
    """Tree of nested connected queues (connected internally and with the provided output queue `Qout`)"""

  @abstractmethod
  def tree(self) -> trees.Tree['Pipeline']:
    """Tree of nested pipelines"""

  def observe(self, get_queue: GetQueue, prefix: tuple[str|int, ...] = ()) -> trees.Tree[ReadQueue]:
    """Tree of nested read queues"""
    return trees.path_map(self.tree(), lambda path, pipe: get_queue(prefix + tuple(path), pipe.Tin))

  @abstractmethod
  def run(self, queues: Q, params: P, /) -> T:
    """Artifacts to run the pipeline"""